"""Azure Blob Storage adapter."""

from __future__ import annotations

import codecs
import dataclasses
import os
import uuid
from collections.abc import Iterable
from datetime import datetime, timedelta, timezone
from typing import Any

from azure.core.exceptions import ResourceNotFoundError
from azure.storage.blob import (
    BlobBlock,
    BlobSasPermissions,
    BlobServiceClient,
    ContainerClient,
    ContentSettings,
    generate_blob_sas,
)
from typing_extensions import override

import file_keeper as fk


@dataclasses.dataclass
class Settings(fk.Settings):
    """Azure Blob Storage settings."""

    account_name: str | None = None
    """Name of the account."""
    account_key: str = ""
    """Key for the account."""

    account_url: str = "https://{account_name}.blob.core.windows.net"
    """Custom resource URL."""
    ## azurite
    # account_url: str = "http://127.0.0.1:10000/{account_name}"

    client: BlobServiceClient = None  # pyright: ignore[reportAssignmentType]
    """Existing storage client."""

    container_name: str = ""
    """Name of the storage container."""

    container: ContainerClient = None  # pyright: ignore[reportAssignmentType]
    """Existing container client."""

    def __post_init__(self, **kwargs: Any):
        super().__post_init__(**kwargs)

        self.path = self.path.strip("/")

        if not self.client:
            if self.account_name:
                self.account_url = self.account_url.format(account_name=self.account_name)
                credential = {
                    "account_name": self.account_name,
                    "account_key": self.account_key,
                }
            elif self.account_key:
                credential = self.account_key
            else:
                credential = None

            self.client = BlobServiceClient(
                self.account_url,
                credential,
            )

        self.account_url = self.client.url.rstrip("/")
        self.account_name = self.client.account_name

        if not self.container:
            self.container = self.client.get_container_client(self.container_name)
        self.container_name = self.container.container_name

        if not self.container.exists():
            if self.initialize:
                self.container.create_container()
            else:
                raise fk.exc.InvalidStorageConfigurationError(
                    self.name,
                    f"container `{self.container_name}` does not exist",
                )


class Uploader(fk.Uploader):
    """Azure Blob Storage uploader.

    Supports single-part and multipart uploads.
    """

    storage: AzureBlobStorage
    capabilities = fk.Capability.CREATE | fk.Capability.MULTIPART

    @override
    def upload(self, location: fk.Location, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)
        blob = self.storage.settings.container.get_blob_client(filepath)

        if not self.storage.settings.override_existing and blob.exists():
            raise fk.exc.ExistingFileError(self.storage, location)

        result = blob.upload_blob(
            upload.stream,
            overwrite=True,
            content_settings=ContentSettings(content_type=upload.content_type),
        )

        return fk.FileData(
            location,
            upload.size,
            upload.content_type,
            codecs.encode(result["content_md5"], "hex").decode(),
        )

    @override
    def multipart_start(self, location: fk.Location, size: int, extras: dict[str, Any]) -> fk.FileData:
        return fk.FileData.from_dict(
            extras,
            location=location,
            size=size,
            storage_data={"multipart": True, "parts": {}, "uploaded": 0},
        )

    @override
    def multipart_refresh(self, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(data.location)
        blob = self.storage.settings.container.get_blob_client(filepath)

        try:
            blocks = blob.get_block_list("uncommitted")[1]
        except ResourceNotFoundError as err:
            raise fk.exc.MissingFileError(self.storage, data.location) from err

        storage_data = {}
        if blocks:
            parts = {}
            uploaded = 0
            for idx, block in enumerate(blocks):
                parts[idx] = block["id"]
                uploaded += block["size"]

            storage_data.update(multipart=True, parts=parts, uploaded=uploaded)
        return fk.FileData.from_object(data, storage_data=storage_data)

    @override
    def multipart_remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        blob_client = self.storage.settings.container.get_blob_client(filepath)
        try:
            blob_client.delete_blob()
        except ResourceNotFoundError:
            return False

        return True

    @override
    def multipart_complete(self, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(data.location)
        blob = self.storage.settings.container.get_blob_client(filepath)

        if data.size != data.storage_data["uploaded"]:
            raise fk.exc.UploadSizeMismatchError(data.storage_data["uploaded"], data.size)

        blob.commit_block_list([BlobBlock(item[1]) for item in sorted(data.storage_data["parts"].items())])

        return fk.FileData.from_object(data, storage_data={})

    @override
    def multipart_update(self, data: fk.FileData, upload: fk.Upload, part: int, extras: dict[str, Any]) -> fk.FileData:
        size = data.storage_data["uploaded"] + upload.size
        if size > data.size:
            raise fk.exc.UploadOutOfBoundError(size, data.size)

        filepath = self.storage.full_path(data.location)
        blob = self.storage.settings.container.get_blob_client(filepath)

        block_id = uuid.uuid4().hex
        blob.stage_block(block_id, upload.stream)

        result = fk.FileData.from_object(data)
        result.storage_data["parts"][part] = block_id
        result.storage_data["uploaded"] = size
        return result


class Reader(fk.Reader):
    """Azure Blob Storage reader.

    Supports streaming and permanent links.
    """

    storage: AzureBlobStorage
    capabilities = fk.Capability.STREAM | fk.Capability.LINK_PERMANENT

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        filepath = self.storage.full_path(data.location)

        blob = self.storage.settings.container.get_blob_client(filepath)
        if not blob.exists():
            raise fk.exc.MissingFileError(self.storage, data.location)

        return blob.download_blob().chunks()

    @override
    def permanent_link(self, data: fk.FileData, extras: dict[str, Any]) -> str:
        account_url = self.storage.settings.account_url
        container = self.storage.settings.container
        filepath = self.storage.full_path(data.location)

        return f"{account_url}/{container.container_name}/{filepath}"


class Manager(fk.Manager):
    """Azure Blob Storage manager."""

    storage: AzureBlobStorage
    capabilities = (
        fk.Capability.REMOVE
        | fk.Capability.SIGNED
        | fk.Capability.EXISTS
        | fk.Capability.SCAN
        | fk.Capability.ANALYZE
        | fk.Capability.COPY
        | fk.Capability.MOVE
    )

    @override
    def copy(self, location: fk.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        src_filepath = self.storage.full_path(data.location)
        blob = self.storage.settings.container.get_blob_client(src_filepath)
        if not blob.exists():
            raise fk.exc.MissingFileError(self.storage, data.location)

        dest_filepath = self.storage.full_path(location)
        dest = self.storage.settings.container.get_blob_client(dest_filepath)
        if not self.storage.settings.override_existing and dest.exists():
            raise fk.exc.ExistingFileError(self.storage, location)

        account_url = self.storage.settings.account_url
        container_name = self.storage.settings.container_name
        url = f"{account_url}/{container_name}/{src_filepath}"
        dest.start_copy_from_url(url)
        return self.analyze(location, extras)

    @override
    def move(self, location: fk.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        self.copy(location, data, extras)
        self.remove(data, extras)
        return self.analyze(location, extras)

    @override
    def analyze(self, location: fk.Location, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)
        blob = self.storage.settings.container.get_blob_client(filepath)
        if not blob.exists():
            raise fk.exc.MissingFileError(self.storage, location)

        info = blob.get_blob_properties()
        content_info = info["content_settings"]
        return fk.FileData(
            location,
            info["size"],
            content_info["content_type"],
            codecs.encode(content_info["content_md5"], "hex").decode(),
        )

    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        blob_client = self.storage.settings.container.get_blob_client(filepath)
        return blob_client.exists()

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        path = self.storage.settings.path
        # do not add slash when path empty, because it will change it from
        # "current directory" to the "root directory"
        if path:
            path = path.rstrip("/") + "/"

        for name in self.storage.settings.container.list_blob_names(name_starts_with=path):
            yield os.path.relpath(name, path)

    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        blob_client = self.storage.settings.container.get_blob_client(filepath)
        if not blob_client.exists():
            return False

        blob_client.delete_blob()
        return True

    @override
    def signed(self, action: fk.types.SignedAction, duration: int, location: fk.Location, extras: dict[str, Any]):
        """Generate a signed URL for a file.

        Upload requires `x-ms-blob-type: BlockBlob` header.
        """
        perms = {}
        if action == "download":
            perms["read"] = True
        elif action == "upload":
            perms["write"] = True
            perms["create"] = True
        elif action == "delete":
            perms["delete"] = True

        client = self.storage.settings.client
        container = self.storage.settings.container
        filepath = self.storage.full_path(location)

        start_time = datetime.now(timezone.utc)
        expiry_time = start_time + timedelta(seconds=duration)

        sas = generate_blob_sas(
            account_name=client.account_name,  # pyright: ignore[reportArgumentType]
            account_key=self.storage.settings.account_key,
            container_name=container.container_name,
            blob_name=filepath,
            permission=BlobSasPermissions(**perms),
            expiry=expiry_time,
        )
        account_url = self.storage.settings.account_url

        return f"{account_url}/{container.container_name}/{filepath}?{sas}"


class AzureBlobStorage(fk.Storage):
    """Azure Blob Storage adapter.

    Uses `azure-storage-blob <https://pypi.org/project/azure-storage-blob/>`_
    package. Make sure to install it first.

    Example configuration:

    ```py
    import file_keeper as fk

    settings = {
        "type": "file_keeper:azure_blob",
        "account_name": "<account_name>",
        "account_key": "<account_key>",
        "container_name": "<container_name>",
        "path": "optional/path/prefix",
        "initialize": True,  # create container if it does not exist
        ## uncomment following line to use azurite running on 10000 port
        "account_url": "http://127.0.0.1:10000/{account_name}",
    }
    storage = fk.make_storage("azure", settings)
    ```

    Note:
    * `account_name` and `account_key` are required unless you provide an existing
      `client` instance.
    * `container_name` is required.
    * `account_url` is optional and defaults to Azure public cloud. You can use it
      to point to a different cloud or a local emulator like azurite.
    * `path` is optional and can be used to prefix all file locations.
    * `initialize` controls whether the container should be created if it does not
      exist. Defaults to `False`.
    """

    settings: Settings
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader



"""Redis adapter."""

from __future__ import annotations

import dataclasses
import os
from collections.abc import Iterable
from io import BytesIO
from typing import IO, Any, ClassVar, cast

import magic
import redis
from typing_extensions import override

import file_keeper as fk

pool = fk.Registry[redis.ConnectionPool]()


@dataclasses.dataclass
class Settings(fk.Settings):
    """Settings for Redis storage."""

    bucket: str = ""
    """Key of the Redis HASH for uploaded objects."""
    redis: redis.Redis = None  # pyright: ignore[reportAssignmentType]
    """Existing redis connection"""

    url: str = ""
    """URL of the Redis DB. Used only if `redis` is empty"""

    _required_options: ClassVar[list[str]] = ["bucket"]

    def __post_init__(self, **kwargs: Any):
        super().__post_init__(**kwargs)

        if self.redis is None:  # pyright: ignore[reportUnnecessaryComparison]
            if self.url not in pool:
                conn = redis.ConnectionPool.from_url(self.url) if self.url else redis.ConnectionPool()
                pool.register(self.url, conn)

            self.redis = redis.Redis(connection_pool=pool[self.url])


class Uploader(fk.Uploader):
    """Redis uploader."""

    storage: RedisStorage
    capabilities: fk.Capability = fk.Capability.CREATE | fk.Capability.RESUMABLE

    @override
    def upload(self, location: fk.types.Location, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)

        cfg = self.storage.settings

        if not cfg.override_existing and cfg.redis.hexists(cfg.bucket, filepath):
            raise fk.exc.ExistingFileError(self.storage, location)

        reader = fk.HashingReader(upload.stream)

        content: Any = reader.read()
        cfg.redis.hset(cfg.bucket, filepath, content)

        return fk.FileData(
            location,
            reader.position,
            upload.content_type,
            reader.get_hash(),
        )

    @override
    def resumable_start(self, location: fk.Location, size: int, extras: dict[str, Any]) -> fk.FileData:
        upload = fk.Upload(
            BytesIO(),
            location,
            0,
            fk.FileData.content_type,
        )
        tmp_result = self.upload(location, upload, extras)

        return fk.FileData.from_dict(
            extras,
            location=tmp_result.location,
            size=size,
            storage_data={"uploaded": 0, "resumable": True},
        )

    @override
    def resumable_refresh(self, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(data.location)

        cfg = self.storage.settings

        if not cfg.redis.hexists(cfg.bucket, filepath):
            raise fk.exc.MissingFileError(self.storage, data.location)

        result = fk.FileData.from_object(data)
        result.storage_data["uploaded"] = cfg.redis.hstrlen(cfg.bucket, filepath)

        if result.storage_data["uploaded"] == result.size:
            return self._resumable_complete(result)

        return result

    @override
    def resumable_remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        return self.storage.remove(data)

    @override
    def resumable_resume(self, data: fk.FileData, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(data.location)

        cfg = self.storage.settings

        if not cfg.redis.hexists(cfg.bucket, filepath):
            raise fk.exc.MissingFileError(self.storage, data.location)

        current: bytes = cfg.redis.hget(cfg.bucket, filepath)  # pyright: ignore[reportAssignmentType]
        size = len(current)

        expected_size = size + upload.size
        if expected_size > data.size:
            raise fk.exc.UploadOutOfBoundError(expected_size, data.size)

        new_content: Any = current + upload.stream.read()
        cfg.redis.hset(cfg.bucket, filepath, new_content)

        result = fk.FileData.from_object(data)
        result.storage_data["uploaded"] = expected_size
        if result.storage_data["uploaded"] == result.size:
            return self._resumable_complete(result)

        return result

    def _resumable_complete(self, data: fk.FileData) -> fk.FileData:
        filepath = self.storage.full_path(data.location)

        cfg = self.storage.settings
        content = cast("bytes | None", cfg.redis.hget(cfg.bucket, filepath))
        if content is None:
            raise fk.exc.MissingFileError(self.storage, data.location)

        reader = fk.HashingReader(BytesIO(content))

        content_type = magic.from_buffer(next(reader, b""), True)
        if data.content_type and content_type != data.content_type:
            raise fk.exc.UploadTypeMismatchError(
                content_type,
                data.content_type,
            )
        reader.exhaust()

        if data.hash and data.hash != reader.get_hash():
            raise fk.exc.UploadHashMismatchError(reader.get_hash(), data.hash)

        result = fk.FileData.from_object(data, content_type=content_type, hash=reader.get_hash())
        result.storage_data.pop("uploaded")
        result.storage_data.pop("resumable")
        return result


class Reader(fk.Reader):
    """Redis reader."""

    storage: RedisStorage
    capabilities: fk.Capability = fk.Capability.STREAM

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> IO[bytes]:
        return BytesIO(self.content(data, extras))

    @override
    def content(self, data: fk.FileData, extras: dict[str, Any]) -> bytes:
        filepath = self.storage.full_path(data.location)

        cfg = self.storage.settings
        content = cast("bytes | None", cfg.redis.hget(cfg.bucket, filepath))
        if content is None:
            raise fk.exc.MissingFileError(self.storage, data.location)

        return content


class Manager(fk.Manager):
    """Redis manager."""

    storage: RedisStorage

    capabilities: fk.Capability = (
        fk.Capability.COPY
        | fk.Capability.MOVE
        | fk.Capability.REMOVE
        | fk.Capability.EXISTS
        | fk.Capability.SCAN
        | fk.Capability.ANALYZE
    )

    @override
    def copy(self, location: fk.types.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        srcpath = self.storage.full_path(data.location)
        destpath = self.storage.full_path(location)

        cfg = self.storage.settings

        if not cfg.redis.hexists(cfg.bucket, srcpath):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not self.storage.settings.override_existing and cfg.redis.hexists(cfg.bucket, destpath):
            raise fk.exc.ExistingFileError(self.storage, location)

        content: Any = cfg.redis.hget(cfg.bucket, srcpath)
        cfg.redis.hset(cfg.bucket, destpath, content)

        return fk.FileData.from_object(data, location=location)

    @override
    def move(self, location: fk.types.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        srcpath = self.storage.full_path(data.location)
        destpath = self.storage.full_path(location)

        cfg = self.storage.settings

        if not cfg.redis.hexists(cfg.bucket, srcpath):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not cfg.override_existing and cfg.redis.hexists(cfg.bucket, destpath):
            raise fk.exc.ExistingFileError(self.storage, location)

        content: Any = cfg.redis.hget(
            cfg.bucket,
            srcpath,
        )
        cfg.redis.hset(cfg.bucket, destpath, content)
        cfg.redis.hdel(cfg.bucket, srcpath)

        return fk.FileData.from_object(data, location=location)

    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)

        cfg = self.storage.settings
        return bool(cfg.redis.hexists(cfg.bucket, filepath))

    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        cfg = self.storage.settings
        result = cfg.redis.hdel(cfg.bucket, filepath)
        return bool(result)

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        path = self.storage.settings.path
        # do not add slash when path empty, because it will change it from
        # "current directory" to the "root directory"
        if path:
            path = path.rstrip("/") + "/"

        cfg = self.storage.settings
        for key in cast("Iterable[bytes]", cfg.redis.hkeys(cfg.bucket)):
            decoded = key.decode()
            if decoded.startswith(path):
                yield os.path.relpath(key.decode(), path)

    @override
    def analyze(self, location: fk.types.Location, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)
        cfg = self.storage.settings
        value: Any = cfg.redis.hget(cfg.bucket, filepath)
        if value is None:
            raise fk.exc.MissingFileError(self.storage, location)

        reader = fk.HashingReader(BytesIO(value))
        content_type = magic.from_buffer(next(reader, b""), True)
        reader.exhaust()

        return fk.FileData(
            location,
            size=reader.position,
            content_type=content_type,
            hash=reader.get_hash(),
        )


class RedisStorage(fk.Storage):
    """Redis storage adapter.

    This adapter uses a Redis HASH to store files. Each file is stored as a field
    in the HASH, where the field name is the file location and the field value is
    the file content.

    Example configuration:

    ```py
    import file_keeper as fk

    settings = {
        "type": "file_keeper:redis",
        "bucket": "my_bucket",
        "url": "redis://localhost:6379/0",
        "override_existing": False,
    }

    storage = fk.make_storage("redis", settings)
    ```

    Note:
    * The `bucket` setting is required and specifies the key of the Redis HASH.
    * The `url` setting is optional. If not provided, it defaults to `redis://localhost:6379/0`.
    * The `override_existing` setting controls whether existing files can be overwritten.

    """

    settings: Settings
    SettingsFactory = Settings

    ReaderFactory = Reader
    ManagerFactory = Manager
    UploaderFactory = Uploader



"""Proxy-storage adapter."""

from __future__ import annotations

import dataclasses
from collections.abc import Mapping
from typing import Any, ClassVar, cast

import file_keeper as fk


@dataclasses.dataclass
class Settings(fk.Settings):
    """Proxy settings."""

    options: dict[str, Any] = cast("dict[str, Any]", dataclasses.field(default_factory=dict))

    storage: fk.Storage = None  # pyright: ignore[reportAssignmentType]

    def __post_init__(self, **kwargs: Any):
        super().__post_init__(**kwargs)

        if not self.storage:
            self.storage = fk.make_storage(self.name, self.options)


@fk.Storage.register
class ProxyStorage:
    """Wrapper for other storages."""

    hidden = True
    ProxySettingsFactory: ClassVar[type[Settings]] = Settings
    proxy_settings: Settings

    def __init__(self, settings: Mapping[str, Any] | Settings, /):
        self.proxy_settings = (  # pyright: ignore[reportAttributeAccessIssue]
            settings if isinstance(settings, Settings) else self.ProxySettingsFactory.from_dict(settings)
        )

    def __getattr__(self, name: str):
        return getattr(self.proxy_settings.storage, name)



"""ZIP adapter."""

from __future__ import annotations

import dataclasses
import os
import zipfile
from collections.abc import Iterable
from typing import Any, Literal

import magic
from typing_extensions import override

import file_keeper as fk

REMOVE_MARKER = b"FILE_KEEPER REMOVED"


def _exists(info: zipfile.ZipInfo):
    """Check if a file exists (i.e., is not marked as removed)."""
    return info.comment != REMOVE_MARKER


def _open(settings: Settings, mode: Literal["a", "r"]):
    """Open the ZIP file."""
    return zipfile.ZipFile(settings.zip_path, "a")


@dataclasses.dataclass()
class Settings(fk.Settings):
    """Settings for a ZIP storage."""

    zip_path: str = ""
    """Path to the ZIP file used for storage."""

    _required_options = ["zip_path"]


class Uploader(fk.Uploader):
    """Service responsible for writing data into a zip storage."""

    storage: ZipStorage
    capabilities = fk.Capability.CREATE

    @override
    def upload(self, location: fk.Location, upload: fk.Upload, extras: dict[str, Any]):
        filepath = self.storage.full_path(location)

        reader = fk.HashingReader(upload.stream)
        with _open(self.storage.settings, "a") as z:
            try:
                info = z.getinfo(filepath)
            except KeyError:
                pass
            else:
                if not self.storage.settings.override_existing and _exists(info):
                    raise fk.exc.ExistingFileError(self.storage, location)

            z.writestr(filepath, reader.read())

        return fk.FileData(
            location,
            upload.size,
            upload.content_type,
            hash=reader.get_hash(),
        )


class Reader(fk.Reader):
    """Service responsible for reading data from the zip storage."""

    storage: ZipStorage
    capabilities = fk.Capability.STREAM

    @override
    def stream(
        self,
        data: fk.FileData,
        extras: dict[str, Any],
    ):
        filepath = self.storage.full_path(data.location)
        with _open(self.storage.settings, "r") as z:
            try:
                info = z.getinfo(filepath)
            except KeyError as err:
                raise fk.exc.MissingFileError(self.storage, data.location) from err

            if not _exists(info):
                raise fk.exc.MissingFileError(self.storage, data.location)

            return [z.read(info)]


class Manager(fk.Manager):
    """Service responsible for managing data in the zip storage."""

    storage: ZipStorage
    capabilities = fk.Capability.REMOVE | fk.Capability.SCAN | fk.Capability.EXISTS | fk.Capability.ANALYZE

    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]):
        filepath = self.storage.full_path(data.location)
        with _open(self.storage.settings, "a") as z:
            try:
                info = z.getinfo(filepath)
            except KeyError:
                return False

            if not _exists(info):
                return False

            info.file_size = 0
            info.comment = REMOVE_MARKER
            z.writestr(info, b"")
        return True

    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        with _open(self.storage.settings, "a") as z:
            try:
                return _exists(z.getinfo(filepath))
            except KeyError:
                return False

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        path = self.storage.settings.path
        # do not add slash when path empty, because it will change it from
        # "current directory" to the "root directory"
        if path:
            path = path.rstrip("/") + "/"

        with _open(self.storage.settings, "a") as z:
            for info in z.infolist():
                if info.filename.startswith(path) and _exists(info):
                    yield os.path.relpath(info.filename, path)

    @override
    def analyze(self, location: fk.Location, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)
        with _open(self.storage.settings, "a") as z:
            try:
                info = z.getinfo(filepath)
            except KeyError as err:
                raise fk.exc.MissingFileError(self.storage, location) from err

            if not _exists(info):
                raise fk.exc.MissingFileError(self.storage, location)

            reader = fk.HashingReader(z.open(info))
            content_type = magic.from_buffer(next(reader, b""), True)
            reader.exhaust()

            return fk.FileData(
                location,
                size=reader.position,
                content_type=content_type,
                hash=reader.get_hash(),
            )


class ZipStorage(fk.Storage):
    """Storage implementation using a ZIP file.

    This storage uses a ZIP file to store files. It supports uploading, reading,
    removing, and scanning files within the ZIP archive.

    Example configuration:

    ```py
    import file_keeper as fk

    settings = {
        "type": "file_keeper:zip",
        "path": "/path/to/storage.zip",
        "override_existing": False,  # Optional, defaults to False
    }

    storage = fk.make_storage("zip", settings)
    ```

    Note:
    * The `path` setting specifies the location of the ZIP file.
    * The `override_existing` setting determines whether existing files should be
      overridden during upload. Defaults to `False`.
    * Removed files are truncated and marked within the ZIP archive, and are not
      physically deleted.
    """

    settings: Settings
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader



"""Filesystem adapter."""

from __future__ import annotations

import dataclasses
import glob
import logging
import os
import shutil
from collections.abc import Iterable
from io import BytesIO
from typing import IO, Any, ClassVar

import magic
from typing_extensions import override

import file_keeper as fk
from file_keeper.core.utils import SAMPLE_SIZE

log = logging.getLogger(__name__)


# --8<-- [start:storage_cfg]
@dataclasses.dataclass()
class Settings(fk.Settings):
    """Settings for FS storage."""

    # --8<-- [end:storage_cfg]
    _required_options: ClassVar[list[str]] = ["path"]

    # --8<-- [start:storage_cfg_post_init]
    def __post_init__(self, **kwargs: Any):
        super().__post_init__(**kwargs)

        if not os.path.exists(self.path):
            if not self.initialize:
                raise fk.exc.InvalidStorageConfigurationError(
                    self.name,
                    f"path `{self.path}` does not exist",
                )

            try:
                os.makedirs(self.path)
            except PermissionError as err:
                raise fk.exc.InvalidStorageConfigurationError(
                    self.name,
                    f"path `{self.path}` is not writable",
                ) from err

    # --8<-- [end:storage_cfg_post_init]


# --8<-- [start:uploader_def]
class Uploader(fk.Uploader):
    """Filesystem uploader."""

    # --8<-- [end:uploader_def]
    storage: FsStorage

    # --8<-- [start:uploader_capability]
    capabilities: fk.Capability = fk.Capability.CREATE | fk.Capability.RESUMABLE
    # --8<-- [end:uploader_capability]

    # --8<-- [start:uploader_method]
    @override
    def upload(self, location: fk.types.Location, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        # --8<-- [end:uploader_method]
        """Upload file to computed location.

        File location is relative the configured `path`. If `Settings.path` is
        not empty and location leads to the place outside of this path, the
        upload will fail with an exception. But if `Settings.path` is empty, no
        restrictions applied to the final location. Consider using combination
        of `storage.prepare_location` with `settings.location_transformers` to
        sanitize the path, like `safe_relative_path`, if you are using
        filesystem-like storage. For cloud providers or DB storages this should
        not be a problem, as they allow using any characters for the name of
        fileobject.

        Raises:
            ExistingFileError: file exists and overrides are not allowed
            LocationError: unallowed usage of subdirectory

        Returns:
            New file data

        """
        # --8<-- [start:uploader_impl_path]
        dest = self.storage.full_path(location)
        # --8<-- [end:uploader_impl_path]

        # --8<-- [start:uploader_impl_check]
        if not self.storage.settings.override_existing and os.path.exists(dest):
            raise fk.exc.ExistingFileError(self.storage, location)
        # --8<-- [end:uploader_impl_check]

        # --8<-- [start:uploader_impl_makedirs]
        os.makedirs(os.path.dirname(dest), exist_ok=True)
        # --8<-- [end:uploader_impl_makedirs]

        # --8<-- [start:uploader_impl_write]
        reader = upload.hashing_reader()
        with open(dest, "wb") as fd:
            for chunk in reader:
                fd.write(chunk)
        # --8<-- [end:uploader_impl_write]

        # --8<-- [start:uploader_impl_result]
        return fk.FileData(
            location,
            os.path.getsize(dest),
            upload.content_type,
            reader.get_hash(),
        )

    # --8<-- [end:uploader_impl_result]

    @override
    def resumable_start(self, location: fk.Location, size: int, extras: dict[str, Any]) -> fk.FileData:
        upload = fk.Upload(BytesIO(), location, 0, fk.FileData.content_type)

        tmp_result = self.upload(location, upload, extras)

        return fk.FileData.from_dict(
            extras,
            location=tmp_result.location,
            size=size,
            storage_data={"resumable": True, "uploaded": 0},
        )

    @override
    def resumable_refresh(self, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(data.location)

        if not os.path.exists(filepath):
            raise fk.exc.MissingFileError(self.storage, data.location)

        result = fk.FileData.from_object(data)
        result.storage_data["uploaded"] = os.path.getsize(filepath)

        if result.storage_data["uploaded"] == result.size:
            result = self._resumable_complete(result)

        return result

    def _resumable_complete(self, data: fk.FileData):
        filepath = self.storage.full_path(data.location)

        with open(filepath, "rb") as src:
            reader = fk.HashingReader(src)
            content_type = magic.from_buffer(next(reader, b""), True)
            if data.content_type and content_type != data.content_type:
                raise fk.exc.UploadTypeMismatchError(
                    content_type,
                    data.content_type,
                )
            reader.exhaust()

        if data.hash and data.hash != reader.get_hash():
            raise fk.exc.UploadHashMismatchError(reader.get_hash(), data.hash)

        result = fk.FileData.from_object(data, hash=reader.get_hash())
        result.storage_data.pop("uploaded")
        result.storage_data.pop("resumable")
        return result

    @override
    def resumable_resume(self, data: fk.FileData, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(data.location)

        if not os.path.exists(filepath):
            raise fk.exc.MissingFileError(self.storage, data.location)

        uploaded = data.storage_data["uploaded"]
        expected_size = uploaded + upload.size

        if expected_size > data.size:
            raise fk.exc.UploadOutOfBoundError(expected_size, data.size)

        filepath = self.storage.full_path(data.location)
        with open(filepath, "rb+") as dest:
            dest.seek(uploaded)
            for chunk in upload.stream:
                dest.write(chunk)

        result = fk.FileData.from_object(data)
        result.storage_data["uploaded"] = os.path.getsize(filepath)

        if result.storage_data["uploaded"] == result.size:
            result = self._resumable_complete(result)

        return result

    @override
    def resumable_remove(self, data: fk.FileData, extras: dict[str, Any]):
        return self.storage.remove(data, **extras)


# --8<-- [start:reader_impl]
class Reader(fk.Reader):
    """Filesystem reader."""

    storage: FsStorage
    capabilities: fk.Capability = fk.Capability.STREAM

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> IO[bytes]:
        filepath = self.storage.full_path(data.location)
        if not os.path.exists(filepath):
            raise fk.exc.MissingFileError(self.storage, data.location)

        return open(filepath, "rb")  # noqa: SIM115


# --8<-- [end:reader_impl]


class Manager(fk.Manager):
    """Filesystem manager."""

    storage: FsStorage
    # --8<-- [start:manager_capabilities]
    capabilities: fk.Capability = (
        fk.Capability.REMOVE
        | fk.Capability.SCAN
        | fk.Capability.EXISTS
        | fk.Capability.ANALYZE
        | fk.Capability.COPY
        | fk.Capability.MOVE
        | fk.Capability.COMPOSE
        | fk.Capability.APPEND
    )
    # --8<-- [end:manager_capabilities]

    # --8<-- [start:manager_compose]
    @override
    def compose(self, location: fk.types.Location, data: Iterable[fk.FileData], extras: dict[str, Any]) -> fk.FileData:
        dest = self.storage.full_path(location)

        if os.path.exists(dest) and not self.storage.settings.override_existing:
            raise fk.exc.ExistingFileError(self.storage, location)

        sources: list[str] = []
        for item in data:
            src = self.storage.full_path(item.location)

            if not os.path.exists(src):
                raise fk.exc.MissingFileError(self.storage, item.location)

            sources.append(src)

        with open(dest, "wb") as to_fd:
            for src in sources:
                with open(src, "rb") as from_fd:
                    shutil.copyfileobj(from_fd, to_fd)

        return self.analyze(location, extras)

    # --8<-- [end:manager_compose]

    # --8<-- [start:manager_append]
    @override
    def append(self, data: fk.FileData, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        dest = self.storage.full_path(data.location)
        if not os.path.exists(dest):
            raise fk.exc.MissingFileError(self.storage, data.location)

        with open(dest, "ab") as fd:
            fd.write(upload.stream.read())

        return self.analyze(data.location, extras)

    # --8<-- [end:manager_append]

    # --8<-- [start:manager_copy]
    @override
    def copy(self, location: fk.types.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        src = self.storage.full_path(data.location)
        dest = self.storage.full_path(location)

        if not os.path.exists(src):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if os.path.exists(dest) and not self.storage.settings.override_existing:
            raise fk.exc.ExistingFileError(self.storage, location)

        os.makedirs(os.path.dirname(dest), exist_ok=True)
        shutil.copy(src, dest)
        return fk.FileData.from_object(data, location=location)

    # --8<-- [end:manager_copy]

    # --8<-- [start:manager_move]
    @override
    def move(self, location: fk.types.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        src = self.storage.full_path(data.location)
        dest = self.storage.full_path(location)

        if not os.path.exists(src):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if os.path.exists(dest):
            if self.storage.settings.override_existing:
                os.remove(dest)
            else:
                raise fk.exc.ExistingFileError(self.storage, location)

        os.makedirs(os.path.dirname(dest), exist_ok=True)
        shutil.move(src, dest)
        return fk.FileData.from_object(data, location=location)

    # --8<-- [end:manager_move]

    # --8<-- [start:manager_exists]
    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        return os.path.exists(filepath)

    # --8<-- [end:manager_exists]

    # --8<-- [start:manager_remove]
    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        if not os.path.exists(filepath):
            return False

        os.remove(filepath)
        return True

    # --8<-- [end:manager_remove]

    # --8<-- [start:manager_scan]
    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        path = self.storage.settings.path
        search_path = os.path.join(path, "**")

        for entry in glob.glob(search_path, recursive=True):
            if not os.path.isfile(entry):
                continue
            yield os.path.relpath(entry, path)

    # --8<-- [end:manager_scan]

    # --8<-- [start:manager_analyze]
    @override
    def analyze(self, location: fk.types.Location, extras: dict[str, Any]) -> fk.FileData:
        return fk.FileData(
            location,
            size=self.size(location, extras),
            content_type=self.content_type(location, extras),
            hash=self.hash(location, extras),
        )

    # --8<-- [end:manager_analyze]

    @override
    def size(self, location: fk.Location, extras: dict[str, Any]) -> int:
        filepath = self.storage.full_path(location)
        if not os.path.exists(filepath):
            raise fk.exc.MissingFileError(self.storage, location)

        return os.path.getsize(filepath)

    @override
    def hash(self, location: fk.Location, extras: dict[str, Any]) -> str:
        filepath = self.storage.full_path(location)
        if not os.path.exists(filepath):
            raise fk.exc.MissingFileError(self.storage, location)

        with open(filepath, "rb") as src:
            reader = fk.HashingReader(src)
            reader.exhaust()
            return reader.get_hash()

    @override
    def content_type(self, location: fk.Location, extras: dict[str, Any]) -> str:
        filepath = self.storage.full_path(location)
        if not os.path.exists(filepath):
            raise fk.exc.MissingFileError(self.storage, location)

        with open(filepath, "rb") as src:
            return magic.from_buffer(src.read(SAMPLE_SIZE), True)


# --8<-- [start:storage]
class FsStorage(fk.Storage):
    """Filesystem storage adapter.

    Stores files on the local filesystem. The `path` setting must be
    configured to point to the base directory where files are stored.

    Example configuration:

    ```py
    import file_keeper as fk

    settings = {
        "type": "file_keeper:fs",
        "path": "/path/to/storage",
        "initialize": True,
        "override_existing": False,
    }
    storage = fk.make_storage("fs", settings)
    ```

    Note:
    * The `path` must be an absolute path.
    * The `path` directory must be writable by the application.
    * The `location` used in file operations is relative to the `path`.
    * If `Storage.path` is not empty, the `location` is validated
      to prevent directory traversal.
    * Consider using combination of `storage.prepare_location` with
      `settings.location_transformers` that further sanitizes the path, like
      `safe_relative_path`.
    * If `initialize` is `True`, the storage will attempt to create the
      directory if it does not exist.
    * If `override_existing` is `False`, operations that would overwrite an
      existing file will raise an `ExistingFileError`.
    """

    settings: Settings

    SettingsFactory = Settings
    UploaderFactory = Uploader
    ReaderFactory = Reader
    ManagerFactory = Manager


# --8<-- [end:storage]



"""Internal utilities of the extension.

Do not use this module outside of the extension and do not import any other
internal module except for config, types and exceptions. Only independent tools
are stored here, to avoid import cycles.

"""

from __future__ import annotations

import abc
import enum
import functools
import hashlib
import io
import itertools
import logging
import re
from collections.abc import Callable, Iterable, Iterator
from typing import Any, Generic

from typing_extensions import ParamSpec, TypeVar, override

from . import types

T = TypeVar("T")
P = ParamSpec("P")

log = logging.getLogger(__name__)

RE_FILESIZE = re.compile(r"^(?P<size>\d+(?:\.\d+)?)\s*(?P<unit>\w*)$")
CHUNK_SIZE = io.DEFAULT_BUFFER_SIZE
SAMPLE_SIZE = 1024 * 2
BINARY_BASE = 1024
SI_BASE = 1000

UNITS = {
    "": 1,
    "b": 1,
    "k": 10**3,
    "kb": 10**3,
    "m": 10**6,
    "mb": 10**6,
    "g": 10**9,
    "gb": 10**9,
    "t": 10**12,
    "p": 10**15,
    "tb": 10**12,
    "kib": 2**10,
    "mib": 2**20,
    "gib": 2**30,
    "tib": 2**40,
    "pib": 2**50,
}


class run_once(Generic[P, T]):  # noqa: N801
    """Decorator that runs function only once and caches the result."""

    EMPTY = object()

    result: T
    func: Callable[P, T]

    def __init__(self, func: Callable[P, T]):
        self.func = func
        self.result = self.EMPTY  # pyright: ignore[reportAttributeAccessIssue]

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        """Run function and cache the result."""
        if self.result is self.EMPTY:
            self.result = self.func(*args, **kwargs)
        return self.result

    def reset(self) -> None:
        """Reset cached result to run function again on next call."""
        self.result = self.EMPTY  # pyright: ignore[reportAttributeAccessIssue]


def ensure_setup(func: Callable[P, T]) -> Callable[P, T]:
    """Initialize file-keeper if required."""

    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        from file_keeper.ext import setup  # noqa: PLC0415

        setup()

        return func(*args, **kwargs)

    return wrapper


class HashingReader:
    """IO stream wrapper that computes content hash while stream is consumed.

    Args:
        stream: iterable of bytes or file-like object
        chunk_size: max number of bytes read at once
        algorithm: hashing algorithm

    Example:
        ```
        reader = HashingReader(readable_stream)
        for chunk in reader:
            ...
        print(f"Hash: {reader.get_hash()}")
        ```
    """

    stream: types.PStream
    chunk_size: int
    algorithm: str
    hashsum: Any
    position: int

    def __init__(
        self,
        stream: types.PStream,
        chunk_size: int = CHUNK_SIZE,
        algorithm: str = "md5",
    ):
        self.hashsum = hashlib.new(algorithm)
        self.stream = stream
        self.chunk_size = chunk_size
        self.algorithm = algorithm
        self.position = 0

    def __iter__(self) -> Iterator[bytes]:
        return self

    def __next__(self) -> bytes:
        chunk = self.stream.read(self.chunk_size)
        if not chunk:
            raise StopIteration

        self.position += len(chunk)
        self.hashsum.update(chunk)
        return chunk

    next: Callable[..., bytes] = __next__

    def read(self) -> bytes:
        """Read and return all bytes from stream at once."""
        return b"".join(self)

    def get_hash(self) -> str:
        """Get current content hash as a string."""
        return self.hashsum.hexdigest()

    def exhaust(self) -> None:
        """Exhaust internal stream to compute final version of content hash.

        Note, this method does not returns data from the stream. The content
        will be irreversibly lost after method execution.
        """
        for _ in self:
            pass


class Capability(enum.Flag):
    """Enumeration of operations supported by the storage.

    Example:
        ```python
        read_and_write = Capability.STREAM | Capability.CREATE
        if storage.supports(read_and_write)
            ...
        ```
    """

    NONE = 0

    ANALYZE = enum.auto()
    """Return file details from the storage."""

    APPEND = enum.auto()
    """Add content to the existing file."""

    COMPOSE = enum.auto()
    """Combine multiple files into a new one in the same storage."""

    COPY = enum.auto()
    """Make a copy of the file inside the same storage."""

    CREATE = enum.auto()
    """Create a file as an atomic object."""

    EXISTS = enum.auto()
    """Check if file exists."""

    LINK_PERMANENT = enum.auto()
    """Make permanent download link."""

    LINK_TEMPORAL = enum.auto()
    """Make expiring download link."""

    LINK_ONE_TIME = enum.auto()
    """Make one-time download link."""

    MOVE = enum.auto()
    """Move file to a different location inside the same storage."""

    MULTIPART = enum.auto()
    """Create file in 3 stages: initialize, upload(repeatable), complete."""

    RANGE = enum.auto()
    """Return specific range of bytes from the file."""

    REMOVE = enum.auto()
    """Remove file from the storage."""

    RESUMABLE = enum.auto()
    """Perform resumable uploads that can be continued after interruption."""

    SCAN = enum.auto()
    """Iterate over all files in the storage."""

    SIGNED = enum.auto()
    """Generate signed URL for specific operation."""

    STREAM = enum.auto()
    """Return file content as stream of bytes."""

    MANAGER_CAPABILITIES = ANALYZE | SCAN | COPY | MOVE | APPEND | COMPOSE | EXISTS | REMOVE | SIGNED
    READER_CAPABILITIES = RANGE | STREAM | LINK_PERMANENT | LINK_TEMPORAL | LINK_ONE_TIME
    UPLOADER_CAPABILITIES = CREATE | MULTIPART | RESUMABLE

    def exclude(self, *capabilities: Capability):
        """Remove capabilities from the cluster.

        Other Args:
            capabilities: removed capabilities

        Example:
            ```python
            cluster = cluster.exclude(Capability.REMOVE)
            ```
        """
        result = Capability(self)
        for capability in capabilities:
            result = result & ~capability
        return result

    def can(self, operation: Capability) -> bool:
        """Check whether the cluster supports given operation."""
        return (self & operation) == operation


def parse_filesize(value: str) -> int:
    """Transform human-readable filesize into an integer.

    Args:
        value: human-readable filesize

    Raises:
        ValueError: size cannot be parsed or uses unknown units

    Example:
        ```python
        size = parse_filesize("10GiB")
        assert size == 10_737_418_240
        ```
    """
    result = RE_FILESIZE.match(value.strip())
    if not result:
        msg = f"Cannot parse filesize: '{value}'"
        raise ValueError(msg)

    size, unit = result.groups()

    multiplier = UNITS.get(unit.lower())
    if not multiplier:
        msg = f"Unknown unit '{unit}' in filesize: '{value}'"
        raise ValueError(msg)

    # using `int` here means that `1.9 bytes` will be truncated to `1 byte`. I
    # don't know situation where fraction of bytes makes sens and this behavior
    # seems sensible
    return int(float(size) * multiplier)


def humanize_filesize(value: int | float, base: int = SI_BASE) -> str:
    """Transform an integer into human-readable filesize.

    Args:
        value: size in bytes
        base: 1000 for SI(KB, MB) or 1024 for binary(KiB, MiB)

    Raises:
        ValueError: base is not recognized

    Example:
        ```python
        size = humanize_filesize(10_737_418_240, base=1024)
        assert size == "10GiB"
        size = humanize_filesize(10_418_240, base=1024)
        assert size == "9.93MiB"
        ```

    """
    if base == SI_BASE:
        suffixes = ["B", "KB", "MB", "GB", "TB", "PB", "EB"]
    elif base == BINARY_BASE:
        suffixes = ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB"]
    else:
        msg = f"Base must be {SI_BASE} (SI) or {BINARY_BASE} (binary), got {base}"
        raise ValueError(msg)

    iteration = 0

    while value >= base:
        iteration += 1
        value /= base

    value = int(value * 100) / 100

    num_format = ".2f" if iteration and not value.is_integer() else ".0f"

    return f"{value:{num_format}}{suffixes[iteration]}"


class AbstractReader(Generic[T], abc.ABC):
    """Abstract wrapper that transforms data into readable stream."""

    source: T
    chunk_size: int

    def __init__(self, source: T, chunk_size: int = CHUNK_SIZE):
        self.source = source
        self.chunk_size = chunk_size

    def __iter__(self):
        while chunk := self.read(self.chunk_size):
            yield chunk

    @abc.abstractmethod
    def read(self, size: int | None = None) -> bytes:
        """Read bytes from the source."""
        ...


class IterableBytesReader(AbstractReader[Iterable[int]]):
    """Wrapper that transforms iterable of bytes into readable stream.

    Example:
        The simplest iterable of bytes is a list that contains byte strings:
        ```py
        parts = [b"hello", b" ", b"world"]
        reader = IterableBytesReader(parts)
        assert reader.read() == b"hello world"
        ```

        More realistic scenario is wrapping generator that produces byte string
        in order to initialize [Upload][file_keeper.Upload]:
        ```py
        def data_generator():
            yield b"hello"
            yield b" "
            yield b"world"

        stream = IterableBytesReader(data_generator())
        upload = Upload(stream, "my_file.txt", 11, "text/plain")
        ```

    """

    def __init__(self, source: Iterable[bytes], chunk_size: int = CHUNK_SIZE):
        super().__init__(itertools.chain.from_iterable(source), chunk_size)

    @override
    def read(self, size: int | None = None):
        return bytes(itertools.islice(self.source, 0, size))



"""Fsspec adapter."""

from __future__ import annotations

import dataclasses
import os
from collections.abc import Iterable
from typing import Any, cast

import fsspec
import magic
from typing_extensions import override

import file_keeper as fk


@dataclasses.dataclass()
class Settings(fk.Settings):
    """Fsspec storage settings."""

    params: dict[str, Any] = cast("dict[str, Any]", dataclasses.field(default_factory=dict))
    """Parameters for fsspec protocol initialization."""
    protocol: str = ""
    """Name of fsspec filesystem protocol."""
    fs: fsspec.AbstractFileSystem = None  # pyright: ignore[reportAssignmentType]
    """Existing fsspec filesystem."""

    def __post_init__(self, **kwargs: Any):
        super().__post_init__(**kwargs)

        if not self.fs:
            if not self.protocol:
                raise fk.exc.MissingStorageConfigurationError(self.name, "protocol")

            try:
                self.fs = fsspec.filesystem(self.protocol, **self.params)
            except ValueError as err:
                raise fk.exc.InvalidStorageConfigurationError(
                    self.name,
                    str(err),
                ) from err


class Uploader(fk.Uploader):
    """FsSpec uploader."""

    storage: FsSpecStorage
    capabilities: fk.Capability = fk.Capability.CREATE

    @override
    def upload(
        self,
        location: fk.types.Location,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        fs = self.storage.settings.fs
        filepath = self.storage.full_path(location)

        if not self.storage.settings.override_existing and fs.exists(filepath):
            raise fk.exc.ExistingFileError(self.storage, location)

        fs.mkdirs(os.path.dirname(filepath), exist_ok=True)
        with fs.open(filepath, "wb") as fobj:
            reader = upload.hashing_reader()
            for chunk in reader:
                fobj.write(chunk)  # pyright: ignore[reportArgumentType]

        return fk.FileData(
            location=location,
            size=reader.position,
            content_type=upload.content_type,
            hash=reader.get_hash(),
        )


class Reader(fk.Reader):
    """FsSpec reader."""

    storage: FsSpecStorage
    capabilities: fk.Capability = fk.Capability.STREAM

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        fs = self.storage.settings.fs
        filepath = self.storage.full_path(data.location)

        try:
            with fs.open(filepath, "rb") as src:  # pyright: ignore[reportUnknownVariableType]
                yield from src  # pyright: ignore[reportReturnType]

        except FileNotFoundError as err:
            raise fk.exc.MissingFileError(self.storage, data.location) from err


class Manager(fk.Manager):
    """FsSpec manager."""

    storage: FsSpecStorage
    capabilities: fk.Capability = (
        fk.Capability.REMOVE
        | fk.Capability.SCAN
        | fk.Capability.EXISTS
        | fk.Capability.COPY
        | fk.Capability.MOVE
        | fk.Capability.ANALYZE
    )

    @override
    def copy(self, location: fk.types.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        fs = self.storage.settings.fs

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not self.storage.settings.override_existing and self.exists(fk.FileData(location), extras):
            raise fk.exc.ExistingFileError(self.storage, location)

        src_location = self.storage.full_path(data.location)
        dest_location = self.storage.full_path(location)

        fs.mkdirs(os.path.dirname(dest_location), exist_ok=True)
        fs.cp(src_location, dest_location)

        return fk.FileData.from_object(data, location=location)

    @override
    def move(self, location: fk.types.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        fs = self.storage.settings.fs

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not self.storage.settings.override_existing and self.exists(fk.FileData(location), extras):
            raise fk.exc.ExistingFileError(self.storage, location)

        src_location = self.storage.full_path(data.location)
        dest_location = self.storage.full_path(location)

        fs.mkdirs(os.path.dirname(dest_location), exist_ok=True)
        fs.mv(src_location, dest_location)

        return fk.FileData.from_object(data, location=location)

    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)

        fs = self.storage.settings.fs
        return fs.exists(filepath)

    @override
    def analyze(self, location: fk.types.Location, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)
        fs = self.storage.settings.fs
        if not fs.exists(filepath):
            raise fk.exc.MissingFileError(self.storage, location)

        src = fs.open(filepath, "rb")  # pyright: ignore[reportUnknownVariableType]
        with src:
            reader = fk.HashingReader(fk.IterableBytesReader(src))  # pyright: ignore[reportArgumentType]
            content_type = magic.from_buffer(next(reader, b""), True)
            reader.exhaust()

        return fk.FileData(location, size=reader.position, content_type=content_type, hash=reader.get_hash())

    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        fs = self.storage.settings.fs

        if not fs.exists(filepath):
            return False

        fs.rm(filepath)
        return True

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        path = self.storage.settings.path
        fs = self.storage.settings.fs

        for name in self._scan(fs, path):
            yield os.path.relpath(name, path)

    def _scan(self, fs: fsspec.AbstractFileSystem, path: str) -> Iterable[str]:
        for entry in fs.ls(path, detail=True):  # pyright: ignore[reportUnknownVariableType]
            if entry["type"] == "file":
                yield entry["name"]
            elif entry["type"] == "directory":
                yield from self._scan(fs, entry["name"])


class FsSpecStorage(fk.Storage):
    """Fsspec storage.

    This storage uses [fsspec](https://filesystem-spec.readthedocs.io/en/latest/)
    to provide a unified interface for various filesystems, including local
    filesystems, cloud storage services (like Amazon S3, Google Cloud Storage,
    Azure Blob Storage), and more.

    The `protocol` parameter specifies the filesystem type (e.g., `s3`, `gcs`,
    `file`, etc.), while the `params` dictionary allows you to pass additional
    configuration options required by the chosen filesystem. For example,
    when using S3, you might need to provide access keys and region information.

    Example configuration:

    ```py
    import file_keeper as fk

    settings = {
        "type": "file_keeper:fsspec",
        "protocol": "local",
        "path": "/tmp/file-keeper",
        "params": {"auto_mkdir": True},
    }

    storage = fk.make_storage("fsspec", settings)
    ```

    Note:
    * Ensure that the necessary dependencies for the chosen filesystem are
      installed. For example, to use S3, you might need to install `s3fs`.
    * The `path` setting in the base `fk.Settings` class is used as a base path
        for relative file operations. Ensure that your `location` values are
        relative to this base path.

    """

    settings: Settings
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader

    @override
    def compute_capabilities(self) -> fk.Capability:
        return super().compute_capabilities()



"""Main entrypoint of the file-keeper.

file-keeper is a Python library that provides a unified interface for
storing, retrieving, and managing files across different storage backends
including local filesystem, cloud services (S3, GCS, Azure), and in-memory
storage.
"""

from .core import exceptions as exc
from .core import types
from .core.data import BaseData, FileData
from .core.registry import Registry
from .core.storage import (
    Manager,
    Reader,
    Settings,
    Storage,
    Uploader,
    adapters,
    get_storage,
    list_adapters,
    make_storage,
)
from .core.types import Location, SignedAction
from .core.upload import Upload, make_upload
from .core.utils import (
    Capability,
    HashingReader,
    IterableBytesReader,
    humanize_filesize,
    parse_filesize,
)

__all__ = [
    "BaseData",
    "Capability",
    "Location",
    "SignedAction",
    "FileData",
    "HashingReader",
    "IterableBytesReader",
    "Manager",
    "Reader",
    "Registry",
    "Settings",
    "Storage",
    "Upload",
    "Uploader",
    "adapters",
    "exc",
    "hookimpl",
    "humanize_filesize",
    "make_storage",
    "get_storage",
    "make_upload",
    "list_adapters",
    "parse_filesize",
    "types",
    "ext",
]

from .ext import hookimpl  # must be the last line to avoid circular imports



"""Google Cloud Storage adapter."""

from __future__ import annotations

import base64
import dataclasses
import os
import re
from collections.abc import Iterable
from datetime import timedelta
from http import HTTPStatus
from typing import Any, cast

import urllib3
from google.api_core.exceptions import Forbidden
from google.auth.credentials import AnonymousCredentials, Credentials
from google.cloud.storage import Blob, Bucket, Client
from google.oauth2.service_account import Credentials as ServiceAccountCredentials
from typing_extensions import override

import file_keeper as fk

RE_RANGE = re.compile(r"bytes=(?P<first_byte>\d+)-(?P<last_byte>\d+)")
HTTP_RESUME = 308


def decode(value: str) -> str:
    """Normalize base64-encoded md5-hash of file content."""
    return base64.decodebytes(value.encode()).hex()


@dataclasses.dataclass()
class Settings(fk.Settings):
    """Settings for GCS adapter."""

    bucket_name: str = ""
    """Name of the storage bucket."""
    client: Client = None  # pyright: ignore[reportAssignmentType]
    """Existing storage client."""
    bucket: Bucket = None  # pyright: ignore[reportAssignmentType]
    """Existing storage bucket."""

    credentials_file: str = ""
    """Path to the JSON with cloud credentials."""
    credentials: Credentials | None = None
    """Existing cloud credentials."""
    project_id: str = ""
    """The project which the client acts on behalf of."""

    client_options: dict[str, Any] | None = None
    """Client options for storage client."""

    def __post_init__(
        self,
        **kwargs: Any,
    ):
        super().__post_init__(**kwargs)

        # GCS ignores first slash and keeping it complicates work for
        # os.path.relpath
        self.path = self.path.strip("/")

        if not self.client:
            if not self.credentials:
                if self.credentials_file:
                    try:
                        self.credentials = ServiceAccountCredentials.from_service_account_file(self.credentials_file)
                    except OSError as err:
                        raise fk.exc.InvalidStorageConfigurationError(
                            self.name,
                            f"file `{self.credentials_file}` does not exist",
                        ) from err
                    if not self.project_id:
                        self.project_id = self.credentials.project_id or ""
                else:
                    self.credentials = AnonymousCredentials()

            if not self.project_id:
                raise fk.exc.MissingStorageConfigurationError(self.name, "project_id")

            self.client = Client(
                self.project_id,
                credentials=self.credentials,
                client_options=self.client_options,
            )

        if not self.bucket:
            self.bucket = self.client.bucket(self.bucket_name)

        if not self.bucket.exists():
            if self.initialize:
                self.client.create_bucket(self.bucket_name)
            else:
                raise fk.exc.InvalidStorageConfigurationError(
                    self.name,
                    f"bucket `{self.bucket_name}` does not exist",
                )


class Uploader(fk.Uploader):
    """GCS Uploader."""

    storage: GoogleCloudStorage

    capabilities: fk.Capability = fk.Capability.CREATE | fk.Capability.RESUMABLE

    @override
    def upload(self, location: fk.types.Location, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)
        bucket = self.storage.settings.bucket
        blob = bucket.blob(filepath)

        if not self.storage.settings.override_existing and blob.exists():
            raise fk.exc.ExistingFileError(self.storage, location)

        blob.upload_from_file(upload.stream, content_type=upload.content_type)

        filehash = decode(blob.md5_hash)
        return fk.FileData(
            location,
            blob.size or upload.size,
            upload.content_type,
            filehash,
        )

    def _resumable_complete(self, data: fk.FileData, info: dict[str, Any]):
        """Finalize a resumable upload session."""
        # size is returned as a string
        size = int(info["size"])
        if data.size and data.size != size:
            raise fk.exc.UploadSizeMismatchError(size, data.size)

        filehash = decode(info["md5Hash"])
        if data.hash and data.hash != filehash:
            raise fk.exc.UploadHashMismatchError(filehash, data.hash)

        content_type = info["contentType"]
        if data.content_type and data.content_type != content_type:
            raise fk.exc.UploadTypeMismatchError(content_type, data.content_type)

        result = fk.FileData.from_object(data, size=size, content_type=content_type, hash=hash)
        result.storage_data.pop("gcs_resumable")
        result.storage_data.pop("resumable")
        return result

    @override
    def resumable_start(self, location: fk.Location, size: int, extras: dict[str, Any]) -> fk.FileData:
        """Start a resumable upload session.

        [Google Cloud
        Documentation](https://cloud.google.com/storage/docs/performing-resumable-uploads#initiate-session)
        contains detailed description of the process.

        Accepts additional extras:

        * origin: If set, the upload can only be completed by a user-agent that
            uploads from the given origin. This can be useful when passing the
            session to a web client.

        """
        filepath = self.storage.full_path(location)
        bucket = self.storage.settings.bucket
        blob = bucket.blob(filepath)

        params = {k: extras[k] for k in ["content_type", "origin"] if k in extras}

        url = cast("str | None", blob.create_resumable_upload_session(size=size, **params))

        if not url:
            msg = "Cannot initialize session URL"
            raise fk.exc.ResumableUploadError(msg)

        return fk.FileData.from_dict(
            extras,
            location=location,
            size=size,
            storage_data={
                "gcs_resumable": {"session_url": url, "uploaded": 0, "origin": extras.get("origin")},
                "resumable": True,
            },
        )

    @override
    def resumable_refresh(self, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:  # noqa: C901
        if not data.storage_data.get("resumable"):
            msg = "Not a resumable data"
            raise fk.exc.StorageDataError(msg)

        session_url: str | None = data.storage_data.get("gcs_resumable", {}).get("session_url")

        if not session_url:
            msg = "`gcs_resumable.session_url` is missing"
            raise fk.exc.StorageDataError(msg)

        timeout = extras.get("request_timeout", 10)
        resp = urllib3.request(
            "PUT",
            session_url,
            headers={
                "content-range": f"bytes */{data.size}",
                "content-length": "0",
            },
            timeout=timeout,
        )

        if resp.status in [HTTPStatus.GONE, HTTPStatus.NOT_FOUND]:
            raise fk.exc.MissingFileError(self.storage, data.location)

        if resp.status in [HTTPStatus.OK, HTTPStatus.CREATED]:
            return self._resumable_complete(data, resp.json())

        if resp.status != HTTPStatus.PERMANENT_REDIRECT:
            msg = f"Unexpected status code {resp.status}"
            raise fk.exc.ResumableUploadError(msg, resp)

        result = fk.FileData.from_object(data)
        if "range" in resp.headers:
            range_match = RE_RANGE.match(resp.headers["range"])
            if not range_match:
                raise fk.exc.ExtrasError(
                    {
                        "session_url": [
                            "Invalid response from Google Cloud:" + " missing range header",
                        ],
                    },
                )
            result.storage_data["gcs_resumable"]["uploaded"] = int(range_match.group("last_byte")) + 1
        else:
            result.storage_data["gcs_resumable"]["uploaded"] = 0

        return data

    @override
    def resumable_resume(self, data: fk.FileData, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        gcs_data = data.storage_data["gcs_resumable"]

        first_byte = gcs_data["uploaded"]
        last_byte = first_byte + upload.size - 1
        size = data.size

        if last_byte >= size:
            raise fk.exc.UploadOutOfBoundError(last_byte, size)

        if last_byte + 1 != size and upload.size < 256 * 1024:
            msg = "Only the final part can be smaller than 256KiB"
            raise fk.exc.ResumableUploadError(msg)

        resp = urllib3.request(
            "PUT",
            gcs_data["session_url"],
            body=upload.stream.read(),
            headers={
                "content-range": f"bytes {first_byte}-{last_byte}/{size}",
            },
            timeout=extras.get("request_timeout", 10),
        )

        if resp.status in [HTTPStatus.OK, HTTPStatus.CREATED]:
            return self._resumable_complete(data, resp.json())

        if resp.status != HTTPStatus.PERMANENT_REDIRECT:
            msg = f"Cannot resume upload: {resp.status} {resp.data}"
            raise fk.exc.ResumableUploadError(msg, resp)

        range_match = RE_RANGE.match(resp.headers["range"])
        if not range_match:
            msg = "Invalid response from Google Cloud"
            raise fk.exc.ResumableUploadError(msg)

        result = fk.FileData.from_object(data)
        result.storage_data["gcs_resumable"]["uploaded"] = int(range_match.group("last_byte")) + 1

        return result

    @override
    def resumable_remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        gcs_data = data.storage_data["gcs_resumable"]
        resp = urllib3.request("DELETE", gcs_data["session_url"])
        cancel_status = 499
        return resp.status == cancel_status


class Reader(fk.Reader):
    """GCS Reader."""

    storage: GoogleCloudStorage

    capabilities = fk.Capability.STREAM | fk.Capability.LINK_PERMANENT

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        name = self.storage.full_path(data.location)
        bucket = self.storage.settings.bucket
        blob = bucket.blob(name)

        if not blob.exists():
            raise fk.exc.MissingFileError(self.storage, data.location)

        with blob.open("rb") as stream:
            yield from cast(Iterable[bytes], stream)

    @override
    def permanent_link(self, data: fk.FileData, extras: dict[str, Any]) -> str:
        name = self.storage.full_path(data.location)
        bucket = self.storage.settings.bucket
        blob = bucket.blob(name)

        if not blob.exists():
            raise fk.exc.MissingFileError(self.storage, data.location)
        return blob.public_url


class Manager(fk.Manager):
    """GCS Manager."""

    storage: GoogleCloudStorage
    capabilities: fk.Capability = (
        fk.Capability.REMOVE
        | fk.Capability.SIGNED
        | fk.Capability.COPY
        | fk.Capability.MOVE
        | fk.Capability.EXISTS
        | fk.Capability.ANALYZE
        | fk.Capability.SCAN
    )

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        path = self.storage.settings.path
        # do not add slash when path empty, because it will change it from
        # "current directory" to the "root directory"
        if path:
            path = path.rstrip("/") + "/"

        bucket = self.storage.settings.bucket

        for blob in cast(Iterable[Blob], bucket.list_blobs(prefix=path)):
            name: str = cast(str, blob.name)
            yield os.path.relpath(name, path)

    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        bucket = self.storage.settings.bucket
        blob = bucket.blob(filepath)
        return blob.exists()

    @override
    def move(self, location: fk.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        src_filepath = self.storage.full_path(data.location)
        dest_filepath = self.storage.full_path(location)

        bucket = self.storage.settings.bucket
        src_blob = bucket.blob(src_filepath)
        if not src_blob.exists():
            raise fk.exc.MissingFileError(self.storage, data.location)

        dest_blob = bucket.blob(dest_filepath)
        if not self.storage.settings.override_existing and dest_blob.exists():
            raise fk.exc.ExistingFileError(self.storage, location)

        bucket.rename_blob(src_blob, dest_filepath)
        return self.analyze(location, extras)

    @override
    def copy(self, location: fk.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        src_filepath = self.storage.full_path(data.location)
        dest_filepath = self.storage.full_path(location)

        bucket = self.storage.settings.bucket
        src_blob = bucket.blob(src_filepath)
        if not src_blob.exists():
            raise fk.exc.MissingFileError(self.storage, data.location)

        dest_blob = bucket.blob(dest_filepath)
        if not self.storage.settings.override_existing and dest_blob.exists():
            raise fk.exc.ExistingFileError(self.storage, location)

        bucket.copy_blob(src_blob, bucket, dest_filepath)
        return self.analyze(location, extras)

    @override
    def analyze(self, location: fk.Location, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)
        bucket = self.storage.settings.bucket
        blob = bucket.blob(filepath)

        if not blob.exists():
            raise fk.exc.MissingFileError(self.storage, location)

        blob.reload()  # pull hash, type, size

        filehash = decode(blob.md5_hash)
        size = cast(int, blob.size)

        return fk.FileData(
            location,
            size,
            blob.content_type,
            filehash,
        )

    @override
    def signed(
        self, action: fk.types.SignedAction, duration: int, location: fk.Location, extras: dict[str, Any]
    ) -> str:
        name = self.storage.full_path(location)

        bucket = self.storage.settings.bucket
        blob = bucket.blob(name)

        method = {"download": "GET", "upload": "PUT", "delete": "DELETE"}[action]

        return blob.generate_signed_url(
            version="v4",
            expiration=timedelta(seconds=duration),
            method=method,
        )

    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        bucket = self.storage.settings.bucket

        blob = bucket.blob(filepath)

        try:
            exists = blob.exists()
        except Forbidden as err:
            raise fk.exc.PermissionError(
                self,
                "exists",
                str(err),
            ) from err

        if exists:
            try:
                blob.delete()
            except Forbidden as err:
                raise fk.exc.PermissionError(
                    self,
                    "remove",
                    str(err),
                ) from err
            return True
        return False


class GoogleCloudStorage(fk.Storage):
    """Google Cloud Storage adapter.

    Example configuration:
    ```py
    import file_keeper as fk

    settings = {
        "type": "file_keeper:gcs",
        "bucket_name": "my-bucket",
        "path": "uploads",
        "credentials_file": "/path/to/credentials.json",
        "project_id": "my-project-id",
        "override_existing": False,
        "initialize": True,
    }

    storage = fk.make_storage("gcss", settings)
    ```

    Note:
    * `credentials_file` or `credentials` along with `project_id` are required
      to initialize the storage client.
    * If `initialize` is set to `True`, the bucket will be created if it
        does not exist.

    """

    settings: Settings
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader



"""Built-in storage adapters."""

import contextlib

from .fs import FsStorage
from .memory import MemoryStorage
from .null import NullStorage
from .proxy import ProxyStorage
from .zip import ZipStorage

AzureBlobStorage = None
with contextlib.suppress(ImportError):
    from .azure_blob import AzureBlobStorage

FilebinStorage = None
with contextlib.suppress(ImportError):
    from .filebin import FilebinStorage


GoogleCloudStorage = None
with contextlib.suppress(ImportError):
    from .gcs import GoogleCloudStorage

LibCloudStorage = None
with contextlib.suppress(ImportError):
    from .libcloud import LibCloudStorage

OpenDalStorage = None
with contextlib.suppress(ImportError):
    from .opendal import OpenDalStorage

RedisStorage = None
with contextlib.suppress(ImportError):
    from .redis import RedisStorage


S3Storage = None
with contextlib.suppress(ImportError):
    from .s3 import S3Storage

FsSpecStorage = None
with contextlib.suppress(ImportError):
    from .fsspec import FsSpecStorage

ObjectStoreStorage = None
with contextlib.suppress(ImportError):
    from .obstore import ObjectStoreStorage

SqlAlchemyStorage = None
with contextlib.suppress(ImportError):
    from .sqlalchemy import SqlAlchemyStorage

__all__ = [
    "AzureBlobStorage",
    "FilebinStorage",
    "FsStorage",
    "GoogleCloudStorage",
    "LibCloudStorage",
    "MemoryStorage",
    "NullStorage",
    "OpenDalStorage",
    "ProxyStorage",
    "RedisStorage",
    "S3Storage",
    "SqlAlchemyStorage",
    "ZipStorage",
    "FsSpecStorage",
    "ObjectStoreStorage",
]
